;#######################################################################
;
; NAME:
;  dm_ghoston
;
; PURPOSE:
;  this program search for ghoston for a triple axis instrument
;  reference: Physica B 350 (2004) 11-16
;  
; CATEGORY:
;  TAS tools
;
; AUTHOR:
;  Yiming Qiu
;  NIST Center for Neutron Research
;  100 Bureau Drive, Gaithersburg, MD 20899-6102
;  United States
;  yiming.qiu@nist.gov
;  October, 2018
;
; LICENSE:
;  The software in this file is written by an employee of
;  National Institute of Standards and Technology
;  as part of the DAVE software project.
;
;  The DAVE software package is not subject to copyright protection
;  and is in the public domain. It should be considered as an
;  experimental neutron scattering data reduction, visualization, and
;  analysis system. As such, the authors assume no responsibility
;  whatsoever for its use, and make no guarantees, expressed or
;  implied, about its quality, reliability, or any other
;  characteristic. The use of certain trade names or commercial
;  products does not imply any endorsement of a particular product,
;  nor does it imply that the named product is necessarily the best
;  product for the stated purpose. We would appreciate acknowledgment
;  if the DAVE software is used or if the code in this file is
;  included in another product.
;
;#######################################################################

@dcs_cryst_align

pro dm_ghoston_Exit,tlb
    widget_control,tlb,get_uvalue=self
    if obj_valid(self) then obj_destroy,self
end

;widget event handler
pro dm_ghoston_event,event
    widget_control,event.handler,get_uvalue=self
    if obj_valid(self) then self->event,event
end

pro dm_ghoston::adjust_E
    ind = where(finite([self.Ei,self.Ef,self.E],/nan),count)
    if count eq 0 then begin
       if self.E ne self.Ei-self.Ef then begin
          if self.fixeief then self.Ei = self.Ef+self.E else self.Ef = self.Ei-self.E
          adjust = 1b
       endif
    endif else if count eq 1 then begin
       case ind[0] of 
            0:  self.Ei = self.Ef+self.E 
            1:  self.Ef = self.Ei-self.E 
            2:  self.E  = self.Ei-self.Ef
       endcase
       adjust = 1b
    endif
    if keyword_set(adjust) then begin
       self->my_widget_control,'eief',set_value=dm_to_string(([self.Ei,self.Ef])[self.fixeief])
       self->my_widget_control,'efei',set_value=dm_to_string(([self.Ef,self.Ei])[self.fixeief])
       self->my_widget_control,'en',set_value=dm_to_string(self.E)
    endif
end

function dm_ghoston::check_parm,nomesg=nomesg
    mesg = ''
    if total(finite(self.lattparm,/nan)) ne 0 then mesg = [mesg,'Missing lattice parameters.']
    if total([finite(self.U,/nan),finite(self.V,/nan)]) ne 0 then mesg = [mesg,'Missing scattering plane vectors.'] $
    else if norm(crossp(self.U,self.V)) eq 0 then mesg = [mesg,'Parallel scattering plane axes U and V.']
    if total(finite([self.Ei,self.Ef,self.E],/nan)) ne 0 then mesg = [mesg,'Missing energy information.'] $
    else begin
       if self.Ei le 0 then mesg = [mesg,'Invalid Ei.']
       if self.Ef le 0 then mesg = [mesg,'Invalid Ef.']
    endelse
    if total(finite(self.Q,/nan)) ne 0 then mesg = [mesg,'Missing Q vector.']
    if n_elements(mesg) eq 1 then begin ;check if Q is on the scattering plane
       ;direct -> reciprocal basis vectors
       self.abc_r = dm_basis_r(self.lattparm,error=error)
       if keyword_set(error) then mesg = [mesg,'Error encountered in calculating the resiprocal lattice parameters.'] $
       else begin
          tmp = crossp(self.U,self.V)
          if transpose(self.Q)#tmp ne 0 then mesg = [mesg,'Q is not in the scattering plane.'] $
          else if norm(self.abc_r#self.Q) gt 2.*!pi/dm_e2lambda(self.Ei)+2.*!pi/dm_e2lambda(self.Ef) then mesg=[mesg,'Q is too large for the give Ei and Ef.']
       endelse
    endif
    if n_elements(mesg) gt 1 then begin ;fatal errors
       if ~keyword_set(nomesg) then ok = dialog_message(mesg[1:*],/error,dialog_parent=self.tlb,/center)
       return,0b
    endif
    if ~self.flags[0] and (total(finite([self.dEi,self.dEf],/nan)) ne 0) and ~keyword_set(nomesg) then begin
       if finite(self.dEi,/nan) then mesg = [mesg,'dEi = '+dm_to_string(self.eresl)+' * Ei is used for the missing Ei resolution.']
       if finite(self.dEf,/nan) then mesg = [mesg,'dEf = '+dm_to_string(self.eresl)+' * Ef is used for the missing Ef resolution.']
       mesg = [mesg,'The factor can be changed in the Options menu.']
       self.flags[0] = 1b
    endif
    if ~keyword_set(nomesg) and (n_elements(mesg) gt 1) then ok = dialog_message(mesg[1:*],dialog_parent=self.tlb,/center)  ;non-fatal error
    return,1b
end

;returns l or r
function dm_ghoston::currenttabnum
    wid = widget_info(self.tlb,find_by_uname='tabbase')
    return, widget_info(wid,/tab_current)
end

pro dm_ghoston::draw,lefthand=lefthand
    l_r = keyword_set(lefthand)
    pobj = self.plots[l_r]
    pobj->erase
    if ~ptr_valid(self.blast[l_r]) then return
    abc_r = self.abc_r
    hkl   = *(self.hkl[l_r])
    blast = *(self.blast[l_r])
    a3a4  = self.a3a4[*,l_r]
    if self.viewsel[l_r] gt 0 then begin
       ind   =  (*self.listind[l_r])[self.viewsel[l_r]]
       ind_f = where(blast eq 0,complement=ind_l)
       case ind of
            -2:   begin ;all bragg first
                  hkl = hkl[*,ind_f]
                  blast = blast[ind_f]
                  end
            -3:   begin ;all bragg last
                  hkl = hkl[*,ind_l]
                  blast = blast[ind_l]
                  end
            else: begin
                  hkl   = hkl[*,ind]
                  blast = blast[ind]
                  end
       endcase
    endif
    ki = self.kihkl[*,l_r] & kf = self.kfhkl[*,l_r]
    if (total(blast) eq 0) or (total(blast) eq n_elements(blast)) then begin
       p0 = ([[ki],[kf]])[*,blast[0]]
       circle = self.flags[3]
    endif else p0 = [0.,0.,0.]
    U = self.U-p0 & V = self.V-p0 & ki = ki-p0 & kf = kf-p0 & p0 = [0.,0,0]-p0 
    for i=0,2 do if max(abs([U[i],V[i],reform(hkl[i,*])])) eq 0 then is2d = i
    ps = float(hkl)
    for i=0,n_elements(blast)-1 do if blast[i] then ps[*,i] = kf+hkl[*,i] else ps[*,i] = ki-hkl[*,i]
    col_step = 30
    cols = ['dark green','red','dark red','dark cyan','blue']
    tmp = self.plots[0]->getpalette(33)    ;Blue-Red col table
    tmp->getproperty,blue=col_blue,green=col_green,red=col_red
    obj_destroy,tmp
    offset = ([0.15,0.1])[n_elements(is2d) ne 0]
    fsize = self.fsize
    xran = [min([p0[0],self.flags[4]?[U[0],V[0]]:0.0,ki[0],kf[0],reform(ps[0,*])]),max([p0[0],self.flags[4]?[U[0],V[0]]:0.0,ki[0],kf[0],reform(ps[0,*])])]
    yran = [min([p0[1],self.flags[4]?[U[1],V[1]]:0.0,ki[1],kf[1],reform(ps[1,*])]),max([p0[1],self.flags[4]?[U[1],V[1]]:0.0,ki[1],kf[1],reform(ps[1,*])])]
    zran = [min([p0[2],self.flags[4]?[U[2],V[2]]:0.0,ki[2],kf[2],reform(ps[2,*])]),max([p0[2],self.flags[4]?[U[2],V[2]]:0.0,ki[2],kf[2],reform(ps[2,*])])]
    if self.flags[4] then cornertxt = 'a3='+dm_to_string(a3a4[0],res=3)+'\deg, a4='+dm_to_string(a3a4[1],res=3)+'\deg'
    if n_elements(blast) eq 1 then title = 'ghoston: q=('+dm_to_string(self.Q-hkl,sep=',')+'), E='+dm_to_string(self.E)+' meV, \tau=('+dm_to_string(hkl,sep=',')+')' else title=''
    if n_elements(is2d) ne 0 then begin
       tmp = where([0,1,2] ne is2d) & xid = ([0,1,2])[tmp[0]] & yid = ([0,1,2])[tmp[1]] & xyid = [xid,yid]
       tmp = [[xran],[yran],[zran]] & xran = tmp[*,xid] & yran = tmp[*,yid]
       pobj->setproperty,xtit=(['H','K','L'])[xid],ytit=(['H','K','L'])[yid],/showxgrid,/showygrid,gridlinestyle='dashed',gridcolor='grey',axisthick=self.thick,/nodraw
       if self.flags[4] then begin
          pobj->add_plot,[p0[xid],U[xid]],[p0[yid],U[yid]],color=cols[0],/addarrow,thick=self.thick
          pobj->add_text,'U',U[xid]+offset*((U[xid] ge p0[xid])?1:-1),U[yid]+offset*((U[yid] ge p0[yid])?1:-1),color=cols[0],normalize=0,fontsize=fsize
          pobj->add_plot,[p0[xid],V[xid]],[p0[yid],V[yid]],color=cols[0],/addarrow,thick=self.thick
          pobj->add_text,'V',V[xid]+offset*((V[xid] ge p0[xid])?1:-1),V[yid]+offset*((V[yid] ge p0[yid])?1:-1),color=cols[0],normalize=0,fontsize=fsize
       endif
       pobj->add_plot,[p0[xid],ki[xid]],[p0[yid],ki[yid]],color=cols[1],/addarrow,thick=self.thick
       pobj->add_text,'ki',ki[xid]+offset*((ki[xid] ge p0[xid])?1:-1),ki[yid]+offset*((ki[yid] ge p0[yid])?1:-1),color=cols[1],normalize=0,fontsize=fsize
       pobj->add_plot,[p0[xid],kf[xid]],[p0[yid],kf[yid]],color=cols[2],/addarrow,thick=self.thick
       pobj->add_text,'kf',kf[xid]+offset*((kf[xid] ge p0[xid])?1:-1),kf[yid]+offset*((kf[yid] ge p0[yid])?1:-1),color=cols[2],normalize=0,fontsize=fsize
       pobj->add_plot,[kf[xid],ki[xid]],[kf[yid],ki[yid]],color=cols[3],/addarrow,thick=self.thick
       pobj->add_text,'Q',mean([ki[xid],kf[xid]])+offset,mean([ki[yid],kf[yid]])+offset,color=cols[3],normalize=0,fontsize=fsize
       for i=0L,n_elements(blast)-1 do begin
           p = ps[*,i]
           if blast[i] then begin
              p1 = (p+kf)/2.0
              pobj->add_arc,p[xyid],kf[xyid],conv_matrix=abc_r,p_origin=p0[xyid],circle=circle,arcs=arcs,linestyle='dashed',color=cols[1+blast[i]],thick=self.thick
           endif else begin
              p1 = (p+ki)/2.0
              pobj->add_arc,p[xyid],ki[xyid],conv_matrix=abc_r,p_origin=p0[xyid],circle=circle,arcs=arcs,linestyle='dashed',color=cols[1+blast[i]],thick=self.thick
           endelse
           orie = atan(float(hkl[yid,i])/hkl[xid,i])/!dtor
           xran = [(xran[0])<(min(arcs[0,*])),(xran[1])>(max(arcs[0,*]))]
           yran = [(yran[0])<(min(arcs[1,*])),(yran[1])>(max(arcs[1,*]))]
           if n_elements(blast) eq 1 then color = cols[4] else color = [col_red[(1+i*col_step) mod 256],col_green[(1+i*col_step) mod 256],col_blue[(1+i*col_step) mod 256]]
           pobj->add_plot,([[kf],[p],[ki]])[xid,*],([[kf],[p],[ki]])[yid,*],color=color,/addarrow,thick=self.thick
           pobj->add_plot,[p0[xid],p[xid]],[p0[yid],p[yid]],color=color,/addarrow,thick=self.thick
           pobj->add_text,'('+dm_to_string(hkl[*,i],sep=',')+')',p1[xid],p1[yid],color=color,normalize=0,fontsize=fsize,orientation=orie,offset=offset,vertical_alignment=0
       endfor
       xs = 0. & ys = 0.
       xran = xran+[-1,1]*offset*2.5& yran = yran+[-1,1]*offset*2.5
       for h=ceil(xran[0]),floor(xran[1]) do begin
           for k=ceil(yran[0]),floor(yran[1]) do begin
               xs = [xs,h] & ys = [ys,k]
           endfor
       endfor
       pobj->add_plot,xs[1:*],ys[1:*],psym='filled circle',color='black',linestyle='no line'
       pobj->setproperty,xran=xran,yran=yran,xtickinterval=1,xtickminor=0,ytickinterval=1,ytickminor=0,cornertxt=cornertxt,title=title
    endif else begin
       pobj->setproperty,xtit='H',ytit='K',ztit='L',/showxgrid,/showygrid,/showzgrid,gridlinestyle='dashed',gridcolor='grey',axisthick=self.thick,/nodraw
       if self.flags[4] then begin
          pobj->add_plot,[p0[0],U[0]],[p0[1],U[1]],[p0[2],U[2]],color=cols[0],/addarrow,thick=self.thick
          pobj->add_text,'U',U[0],U[1],U[2]+offset,color=cols[0],normalize=0,fontsize=fsize
          pobj->add_plot,[p0[0],V[0]],[p0[1],V[1]],[p0[2],V[2]],color=cols[0],/addarrow,thick=self.thick
          pobj->add_text,'V',V[0],V[1],V[2]+offset,color=cols[0],normalize=0,fontsize=fsize
       endif
       pobj->add_plot,[p0[0],ki[0]],[p0[1],ki[1]],[p0[2],ki[2]],color=cols[1],/addarrow,thick=self.thick
       pobj->add_text,'ki',ki[0],ki[1],ki[2]+offset,color=cols[1],normalize=0,fontsize=fsize
       pobj->add_plot,[p0[0],kf[0]],[p0[1],kf[1]],[p0[2],kf[2]],color=cols[2],/addarrow,thick=self.thick
       pobj->add_text,'kf',kf[0],kf[1],kf[2]+offset,color=cols[2],normalize=0,fontsize=fsize
       pobj->add_plot,[kf[0],ki[0]],[kf[1],ki[1]],[kf[2],ki[2]],color=cols[3],/addarrow,thick=self.thick
       pobj->add_text,'Q',mean([ki[0],kf[0]]),mean([ki[1],kf[1]]),mean([ki[2],kf[2]])+offset,color=cols[3],normalize=0,fontsize=fsize
       for i=0L,n_elements(blast)-1 do begin
           p = ps[*,i]
           if blast[i] then begin
              p1 = (p+kf)/2.0
              pobj->add_arc,p,kf,conv_matrix=abc_r,p_origin=p0,circle=circle,arcs=arcs,linestyle='dashed',color=cols[1+blast[i]],thick=self.thick
           endif else begin
              p1 = (p+ki)/2.0
              pobj->add_arc,p,ki,conv_matrix=abc_r,p_origin=p0,circle=circle,arcs=arcs,linestyle='dashed',color=cols[1+blast[i]],thick=self.thick
           endelse
           baseline = hkl[*,i]/norm(hkl[*,i])
           if abs(atan(baseline[1],baseline[0])) gt !pi/2 then baseline = -baseline
           updir = crossp([-baseline[1],baseline[0],0],baseline)
           if updir[2] lt 0 then updir = -updir
           xran = [(xran[0])<(min(arcs[0,*])),(xran[1])>(max(arcs[0,*]))]
           yran = [(yran[0])<(min(arcs[1,*])),(yran[1])>(max(arcs[1,*]))]
           zran = [(zran[0])<(min(arcs[2,*])),(zran[1])>(max(arcs[2,*]))]
           if n_elements(blast) eq 1 then color = cols[4] else color = [col_red[(1+i*col_step) mod 256],col_green[(1+i*col_step) mod 256],col_blue[(1+i*col_step) mod 256]]
           pobj->add_plot,([[kf],[p],[ki]])[0,*],([[kf],[p],[ki]])[1,*],([[kf],[p],[ki]])[2,*],color=color,/addarrow,thick=self.thick
           pobj->add_plot,[p0[0],p[0]],[p0[1],p[1]],[p0[2],p[2]],color=color,/addarrow,thick=self.thick
           pobj->add_text,'('+dm_to_string(hkl[*,i],sep=',')+')',p1[0],p1[1],p1[2],color=color,normalize=0,fontsize=fsize,baseline=baseline,updir=updir,vertical_alignment=0,offset=offset
       endfor
       xs = 0. & ys = 0. & zs = 0.
       xran = xran+[-1,1]*offset*1.5 & yran = yran+[-1,1]*offset*1.5 & zran = zran+[-0.1,1]*offset*1.5
       for h=ceil(xran[0]),floor(xran[1]) do begin
           for k=ceil(yran[0]),floor(yran[1]) do begin
               for l=ceil(zran[0]),floor(zran[1]) do begin
                   xs = [xs,h] & ys = [ys,k] & zs = [zs,l]
               endfor
           endfor
       endfor
       pobj->add_plot,xs[1:*],ys[1:*],zs[1:*],psym='sphere',symsize=0.01,color='black',linestyle='no line'       
       pobj->setproperty,xran=xran,yran=yran,zran=zran,xtickinterval=1,xtickminor=0,ytickinterval=1,ytickminor=0,ztickinterval=1,ztickminor=0,$
             cornertxt=cornertxt,title=title
    endelse
end

pro dm_ghoston::findhkl,lefthand=lefthand
    l_r = keyword_set(lefthand)
    if finite(self.dEi,/nan) then dEi = self.eresl*self.Ei else dEi = abs(self.dEi)
    if finite(self.dEf,/nan) then dEf = self.eresl*self.Ef else dEf = abs(self.dEf)
    self.kiran = [2.*!pi/dm_e2lambda(self.Ei-dEi),2.*!pi/dm_e2lambda(self.Ei+dEi)]
    self.kfran = [2.*!pi/dm_e2lambda(self.Ef-dEf),2.*!pi/dm_e2lambda(self.Ef+dEf)]
    maxk = (self.kiran[1]>self.kfran[1])
    a3a4 = dcs_cryst_align_a3a4(self.Ei,self.Ef,self.Q,lattparm=self.lattparm,U=self.U,V=self.V,lefthand=lefthand,kihkl=kihkl,kfhkl=kfhkl,error=error)
    if keyword_set(error) then return   ;error encountered, no further calcuation
    self.a3a4[*,l_r]  = a3a4
    self.kihkl[*,l_r] = kihkl
    self.kfhkl[*,l_r] = kfhkl
    hkl = [0,0,0] & bragglast = 0b
    for h=-ceil(maxk/norm(self.abc_r[*,0])*2)-1,ceil(maxk/norm(self.abc_r[*,0]))*2 do begin
        for k=-ceil(maxk/norm(self.abc_r[*,1]))*2-1,ceil(maxk/norm(self.abc_r[*,1]))*2 do begin
            for l=-ceil(maxk/norm(self.abc_r[*,2]))*2-1,ceil(maxk/norm(self.abc_r[*,2]))*2 do begin
                if abs(h)+abs(k)+abs(l) eq 0 then continue
                skip  = 0b
                eo_s  = abs(h+k+l) mod 2
                eo_s4 = abs(h+k+l) mod 4
                s_eo  = total(abs([h,k,l]) mod 2)
                case self.flags[2] of 
                     1: if eo_s eq 1 then skip = 1b                                                        ;bcc
                     2: if (s_eo eq 1) or (s_eo eq 2) then skip = 1b                                       ;fcc
                     3: if (s_eo eq 1) or (s_eo eq 2) or ((s_eo eq 0) and (eo_s4 ne 0)) then skip = 1b     ;diamond fcc
                     4: if ((abs(l) mod 2) eq 1) and ((abs(h+2*k) mod 3) eq 0) then skip = 1b              ;hcp
                     else:
                endcase
                if keyword_set(skip) then continue
                ;bragg first
                tmp = norm(self.abc_r#(kihkl-[h,k,l]))
                if (tmp ge self.kiran[0]) and (tmp le self.kiran[1]) then begin
                   tmp = norm(self.abc_r#(self.Q-[h,k,l]))
                   if (tmp lt self.kiran[1]+self.kfran[1]) and (self.kiran[0] lt self.kfran[1]+tmp) and (self.kfran[0] lt self.kiran[1]+tmp) then begin
                      hkl = [[hkl],[[h,k,l]]]
                      bragglast = [bragglast,0b]
                   endif
                endif
                ;bragg last
                tmp = norm(self.abc_r#(kfhkl+[h,k,l]))
                if (tmp ge self.kfran[0]) and (tmp le self.kfran[1]) then begin
                   tmp = norm(self.abc_r#(self.Q-[h,k,l]))
                   if (tmp lt self.kiran[1]+self.kfran[1]) and (self.kiran[0] lt self.kfran[1]+tmp) and (self.kfran[0] lt self.kiran[1]+tmp) then begin
                      hkl = [[hkl],[[h,k,l]]]
                      bragglast = [bragglast,1b]
                   endif
                endif
            endfor
        endfor
    endfor
    ptr_free,self.hkl[l_r],self.blast[l_r],self.listind[l_r]
    list = 'view all'
    listind = -1
    self.viewsel[l_r] = 0
    if n_elements(bragglast) gt 1 then begin
       bragglast = bragglast[1:*]
       hkl = hkl[*,1:*]
       tmp = dblarr(n_elements(bragglast))
       for i=0,n_elements(bragglast)-1 do tmp[i] = bragglast[i]*1d9+norm(self.abc_r#hkl[*,i])*1d4+hkl[0,i]*1d2+hkl[1,i]*1d1+hkl[2,i]
       ind = sort(tmp)
       self.hkl[l_r] = ptr_new(hkl[*,ind])
       self.blast[l_r] = ptr_new(bragglast[ind])
       tmp = where(bragglast eq 0,count1,ncomplement=count2)
       if ((count1 gt 1) or (count2 gt 1)) and (max([count1,count2]) ne n_elements(bragglast))  then begin
          if count1 ne 0 then begin
             list = [list,'view all bragg first']
             listind = [listind,-2]
          endif
          if count2 ne 0 then begin
             list = [list,'view all bragg last']
             listind = [listind,-3]
          endif
       endif
       for i=0,n_elements(ind)-1 do begin
           list = [list,'bragg '+([' first',' last'])[bragglast[ind[i]]]+' :  ('+dm_to_string(hkl[*,ind[i]],sep=',')+')']
           listind = [listind,i]
       endfor
       self.listind[l_r] = ptr_new(listind)
    endif
    dm_set_droplist,widget_info(self.tlb,find_by_uname='viewsel_'+(['r','l'])[l_r]),select=0,set_value=list,sensitive=ptr_valid(self.blast[l_r])
    if l_r eq self->currenttabnum() then self->my_widget_control,['save','print'],sensitive=ptr_valid(self.blast[l_r])
    return
end

function dm_ghoston::getsymbolfont,fsize
    fonts=['symbol*',$
           '-adobe-symbol-medium-r-normal--14-140-75-75-p-85-adobe-fontspecific',$
           '-adobe-symbol-medium-r-normal--14-100-100-100-p-85-adobe-fontspecific']
    fonts[0] = fonts[0]+dm_to_string(fsize+2,/int)
    catch, myerror
    if myerror ne 0 then return,fonts[0]
    result = fonts[0]
    switch !version.os_family of
           'Windows':break
           'unix':
           'MacOS':begin
                   spawn,'xlsfonts',result,errresult,exit_status=error
                   if error ne 0 then begin
                      old_win = !d.window
                      device,get_fontnames=result,set_font='*'    ;will pop up a window, reason for using xlsfonts first
                      if !d.window ne old_win then wdelete, !d.window
                   endif
                   nlist = n_elements(result)
                   tmp   = bytarr(nlist)
                   for i=0,nlist-1 do tmp[i] = ((strpos(strlowcase(result[i]),'symbol'))<0)+1
                   index = where(tmp,count)
                   if count ne 0 then f_list=result[index]
                   nlist = n_elements(f_list)
                   if nlist ne 0 then begin
                      tmp = intarr(nlist)
                      for i=0,nlist-1 do tmp[i]=dm_to_number((strsplit(f_list[i],'[A-Za-z-]',/regex,/extract))[0],/int)
                      void = min(abs(tmp-fsize+2),index)
                      fonts[1] = f_list[index[0]]
                   endif
                   result = fonts[1]
                   break
                   end
           else:
    endswitch
    return,result
end

;load parameter file
pro dm_ghoston::load_parm,parmfile=parmfile,init=init
    if n_elements(parmfile) eq 0 then parmfile = dm_define_pointer(/gettempdir)+dm_define_pointer(/getpathsep)+'.ghoston'
    openr,unit,parmfile,/get_lun,error=openerr
    if openerr ne 0 then return 
    tmp = ''
    while(~ eof(unit)) do begin
        readf,unit,tmp
        tmp  = strtrim(tmp,2)
        head = strmid(tmp,0,7)
        tmp  = strmid(tmp,7,strlen(tmp)-7)
        case strlowcase(head) of
             'lattpm:': self.lattparm[*] = dm_to_number(strsplit(tmp,'&',/extract,/preserve))
             'orie_u:': self.U[*]        = dm_to_number(strsplit(tmp,'&',/extract,/preserve))
             'orie_v:': self.V[*]        = dm_to_number(strsplit(tmp,'&',/extract,/preserve))
             'fixeif:': self.fixeief     = dm_to_number(tmp,/int)
             'e_resl:': self.eresl       = abs(dm_to_number(tmp))
             'ei_val:': self.Ei          = dm_to_number(tmp)
             'ei_res:': self.dEi         = dm_to_number(tmp)
             'ef_val:': self.Ef          = dm_to_number(tmp)
             'ei_res:': self.dEf         = dm_to_number(tmp)
             'en_val:': self.E           = dm_to_number(tmp)
             'q_vect:': self.Q[*]        = dm_to_number(strsplit(tmp,'&',/extract,/preserve))
             'wrkdir:': self.workdir     = tmp
             else:
        endcase
    endwhile
    free_lun,unit
    if keyword_set(init) then return
    uname = ['latt'+['A','B','C','AA','BB','CC'],'eief','deief','efei','defei','en','orie'+['Ux','Uy','Uz','Vx','Vy','Vz'],'Qx','Qy','Qz']
    value = [self.lattparm,([self.Ei,self.Ef])[self.fixeief],([self.dEi,self.dEf])[self.fixeief],([self.Ef,self.Ei])[self.fixeief],([self.dEi,self.dEf])[self.fixeief],self.E,self.U,self.V ,self.Q]
    for i=0,n_elements(uname)-1 do self->my_widget_control,uname[i],set_value=dm_to_string(value[i])
    dm_set_droplist,widget_info(self.tlb,find_by_uname='fixEiEf'),select=self.fixeief
    self->my_widget_control,'lab_efei',set_value=(['Ef','Ei'])[self.fixeief]
    self->my_widget_control,'lab_deief',set_value=(['dEi','dEf'])[self.fixeief]
    self->my_widget_control,'lab_defei',set_value=(['dEf','dEi'])[self.fixeief]
    if file_test(self.workdir,/directory,/write) then for i=0,1 do self.plots[i]->setproperty,path=self.workdir
end

;save parameter file
pro dm_ghoston::save_parm,parmfile=parmfile
    if n_elements(parmfile) eq 0 then parmfile = dm_define_pointer(/gettempdir)+dm_define_pointer(/getpathsep)+'.ghoston'
    openw,unit,parmfile,/get_lun,error=openerr
    if openerr ne 0 then return
    printf,unit,'lattpm:'+dm_to_string(self.lattparm,sep='&')
    printf,unit,'orie_u:'+dm_to_string(self.U,sep='&')
    printf,unit,'orie_v:'+dm_to_string(self.V,sep='&')
    printf,unit,'fixeif:'+dm_to_string(self.fixeief)
    printf,unit,'e_resl:'+dm_to_string(self.eresl)
    printf,unit,'ei_val:'+dm_to_string(self.Ei)
    printf,unit,'ei_res:'+dm_to_string(self.dEi)
    printf,unit,'ef_val:'+dm_to_string(self.Ef)
    printf,unit,'ef_res:'+dm_to_string(self.dEf)
    printf,unit,'en_val:'+dm_to_string(self.E)
    printf,unit,'q_vect:'+dm_to_string(self.Q,sep='&')
    printf,unit,'wrkdir:'+self.workdir
    free_lun,unit
end

;a widget_control use uname instead of id as identifier
pro dm_ghoston::my_widget_control,uname,base,_ref_extra=extra
    if n_elements(base) eq 0 then base=self.tlb
    if ~widget_info(base,/valid_id) then base=self.tlb
    if ~widget_info(base,/valid_id) then return
    if n_elements(extra) ne 0 then begin
       for i=0,n_elements(uname)-1 do begin
           wid = widget_info(base,find_by_uname=uname[i])
           if wid ne 0 then widget_control,wid,_extra=extra
       endfor
    endif
end

;widget_tex with tab_mode=1,/editable,/all_events,/kbrd_focus_events
function dm_ghoston::my_widget_text,base,ysize=ysize,_ref_extra=extra
    if n_elements(ysize) eq 0 then ysize=1
    return,widget_text(base,tab_mode=1,/editable,/all_events,/kbrd_focus_events,ysize=ysize,_extra=extra)
end

;object event handler
pro dm_ghoston::event,event
    compile_opt IDL2    ;idl2=defint32,strictarr

    ;catch and ignore all errors in this program
    catch, myerror
    if myerror ne 0 then begin
       ok = dialog_message(dialog_parent=self.tlb,!error_state.msg,/error,/center)
       catch,/cancel
       return
    end
    
    widget_type = widget_info(event.id,/type)
    case widget_type of
         1: tmp_value = event.select                                ;button
         3: begin                                                   ;text box
            widget_control,event.id,get_value=tmp_value
            tmp_value = strtrim(tmp_value,2)
            end
         8: tmp_value = widget_info(event.id,/droplist_select)      ;drop list
         12:begin                                                   ;combobox
            tmp_value = event.index
            widget_control,event.id,get_value=tmp_list
            if tmp_list[tmp_value] ne event.str then return         ;to avoid events from set_value in unix operating system
            end
         else:
    endcase
    if n_elements(tmp_value) eq 1 then tmp_value=tmp_value[0]
    uname = widget_info(event.id,/uname)
    type  = tag_names(event,/structure)
    case strlowcase(type) of
         'widget_kbrd_focus':   if widget_info(event.id,/name) eq 'TEXT' then uname = 'text_focus'
         'widget_text_sel':     if self.txtfocusevent then begin
                                   tmp = widget_event(event.id,/nowait)
                                   widget_control,event.id,get_value=tmp
                                   if n_elements(tmp) eq 1 then begin
                                      if strlen(strtrim(tmp,2)) ne 0 then widget_control,event.id,set_text_select=[0,strlen(tmp[0])]
                                   endif   
                                endif
         'dm_plot_changetitle': uname = type
         'dm_plot_isotropic':   uname = type
         else: 
    endcase
    if uname ne 'text_focus' then WIDGET_CONTROL,/HOURGLASS
    self.txtfocusevent = 0b
    case strlowcase(uname) of
         'tlb':        if strlowcase(tag_names(event,/structure)) eq 'widget_base' then begin ;resize event
                       if (event.x lt self.xmin) or (event.y lt self.ymin) then begin
                          event.x = (event.x)>(self.xmin)
                          event.y = (event.y)>(self.ymin)
                       endif
                       event1 = event
                       event1.x = event1.x-self.xoffset
                       event1.y = event1.y-self.yoffset
                       self->my_widget_control,'tabbase',scr_xsize=event1.x+self.xmargin,scr_ysize=event1.y+self.ymargin
                       for i=0,1 do self.plots[i]->resize,event1
                       widget_control,self.tlb,/update,/realize
                       endif     
         'tabbase':    self->my_widget_control,['save','print'],sensitive=ptr_valid(self.blast[self->currenttabnum()])
         'save':       self.plots[self->currenttabnum()]->saveas,'png'
         'print':      self.plots[self->currenttabnum()]->saveas,'printer'  
         'saveparm':   begin
                       path = self.workdir
                       parmfile = dm_choose_file('prm',dialog_parent=self.tlb,/write,path=path)
                       if strlen(parmfile) ne 0 then begin
                          if file_test(path,/directory,/write) then begin
                             self.workdir = path
                             for i=0,1 do self.plots[i]->setproperty,path=path
                          endif
                          self->save_parm,parmfile=parmfile
                       endif
                       end
         'loadparm':   begin
                       path = self.workdir
                       parmfile = dm_choose_file('prm',dialog_parent=self.tlb,/read,path=path)
                       if strlen(parmfile) ne 0 then self->load_parm,parmfile=parmfile
                       end                         
         'exit':       begin
                       widget_control,self.tlb,/destroy
                       return
                       end
         'isoscale':   begin
                       self.flags[1] = ~self.flags[1]
                       dm_set_button,event.id,self.flags[1]
                       for i=0,1 do self.plots[i]->setproperty,isotropic=self.flags[1],/noevent 
                       end 
         'circle':     begin
                       self.flags[3] = ~self.flags[3]
                       dm_set_button,event.id,self.flags[3]
                       for lefthand=0,1 do self->draw,lefthand=lefthand
                       end       
         'showuv':     begin
                       self.flags[4] = ~self.flags[4]
                       dm_set_button,event.id,self.flags[4]
                       for lefthand=0,1 do self->draw,lefthand=lefthand
                       end           
         'fsize':      begin
                       new = dm_dialog_input(['font size:'],xsize=100,default=self.fsize,dialog_parent=self.tlb,cancel=cancel,info='The font size affects all texts in the plots.')
                       if (~keyword_set(cancel)) and finite(new) then begin
                          self.fsize = round(abs(new))
                          for i=0,1 do self.plots[i]->change_font,fsize=self.fsize,/all
                       endif
                       end  
         'thick':      begin
                       new = dm_dialog_input(['line thickness:'],xsize=100,default=self.thick,dialog_parent=self.tlb,cancel=cancel,info='The line thickness affects axes and plots.')
                       if (~keyword_set(cancel)) and finite(new) and (new ne self.thick) then begin 
                          self.thick = abs(new)
                          for lefthand=0,1 do self->draw,lefthand=lefthand
                       endif
                       end                                      
         'dm_plot_isotropic': begin
                       self.flags[1] = event.value
                       dm_set_button,widget_info(self.tlb,find_by_uname='isoscale'),event.value
                       for i=0,1 do begin
                           self.plots[i]->getproperty,isotropic=isotropic
                           if isotropic ne self.flags[1] then self.plots[i]->setproperty,isotropic=self.flags[1],/noevent
                       endfor
                       end
         'eresl':      begin
                       new = dm_dialog_input('factor:',title='E Resolution factor',dialog_parent=self.tlb,default=self.eresl,info='dEi(f) = factor*Ei(f) will be used if dEi(f) is not specified.',/float,cancel=cancel)
                       if keyword_set(cancel) or finite(new,/nan) then return
                       self.eresl = abs(new)
                       recal = 1b
                       end              
         'text_focus': begin
                       if event.enter then self.txtfocusevent = 1b $
                       else widget_control,event.id,set_text_select=[0,0]
                       end
         'latta':      self.lattparm[0] = dm_to_number(tmp_value,/float)   ;a
         'lattb':      self.lattparm[1] = dm_to_number(tmp_value,/float)   ;b
         'lattc':      self.lattparm[2] = dm_to_number(tmp_value,/float)   ;c
         'lattaa':     self.lattparm[3] = dm_to_number(tmp_value,/float)   ;alpha
         'lattbb':     self.lattparm[4] = dm_to_number(tmp_value,/float)   ;beta
         'lattcc':     self.lattparm[5] = dm_to_number(tmp_value,/float)   ;gama
         'orieux':     self.U[0]        = dm_to_number(tmp_value,/float)   ;ux
         'orieuy':     self.U[1]        = dm_to_number(tmp_value,/float)   ;uy
         'orieuz':     self.U[2]        = dm_to_number(tmp_value,/float)   ;uz
         'orievx':     self.V[0]        = dm_to_number(tmp_value,/float)   ;vx
         'orievy':     self.V[1]        = dm_to_number(tmp_value,/float)   ;vy
         'orievz':     self.V[2]        = dm_to_number(tmp_value,/float)   ;vz  
         'qx':         self.Q[0]        = dm_to_number(tmp_value,/float)   ;Qx
         'qy':         self.Q[1]        = dm_to_number(tmp_value,/float)   ;Qx
         'qz':         self.Q[2]        = dm_to_number(tmp_value,/float)   ;Qx
         'fixeief':    if self.fixeief ne tmp_value then begin
                       self.fixeief = tmp_value
                       self->my_widget_control,'lab_efei',set_value=(['Ef','Ei'])[self.fixeief]
                       self->my_widget_control,'lab_deief',set_value=(['dEi','dEf'])[self.fixeief]
                       self->my_widget_control,'lab_defei',set_value=(['dEf','dEi'])[self.fixeief]
                       tmp = self.Ei  & self.Ei  = self.Ef  & self.Ef  = tmp
                       tmp = self.dEi & self.dEi = self.dEf & self.dEf = tmp                                   
                       self->adjust_E
                       end
         'eief':       begin
                       if self.fixeief then begin
                          self.Ef = dm_to_number(tmp_value,/float) 
                       endif else begin
                          self.Ei = dm_to_number(tmp_value,/float)
                       endelse
                       if event.type eq 0 then if event.ch eq 10b then self->adjust_E
                       end
         'deief':      begin
                       if self.fixeief then begin
                          self.dEf = dm_to_number(tmp_value,/float)
                       endif else begin
                          self.dEi = dm_to_number(tmp_value,/float)
                       endelse
                       end              
         'efei':       begin
                       if self.fixeief then begin
                          self.Ei = dm_to_number(tmp_value,/float)
                       endif else begin
                          self.Ef = dm_to_number(tmp_value,/float)
                       endelse
                       if event.type eq 0 then if event.ch eq 10b then self->adjust_E
                       end
         'defei':      begin
                       if self.fixeief then begin
                          self.dEi = dm_to_number(tmp_value,/float)
                       endif else begin
                          self.dEf = dm_to_number(tmp_value,/float)
                       endelse
                       end              
         'en':         begin
                       self.E = dm_to_number(tmp_value,/float)
                       if event.type eq 0 then if event.ch eq 10b then self->adjust_E
                       end  
         'calc':       begin
                       self->adjust_E
                       if self->check_parm() then recal = 1b  ;actual calculation done at the end with recal keyword set
                       end
         'viewsel_r':  if self.viewsel[0] ne tmp_value then begin
                       self.viewsel[0] = tmp_value
                       if ptr_valid(self.blast[0]) then self->draw
                       end
         'viewsel_l':  if self.viewsel[1] ne tmp_value then begin
                       self.viewsel[1] = tmp_value
                       if ptr_valid(self.blast[1]) then self->draw,/lefthand
                       end
         'latt_bcc':   begin
                       self.flags[2] = (self.flags[2] eq 1)?0:1
                       widget_control,event.id,set_button=(self.flags[2] gt 0)
                       self->my_widget_control,'latt_'+['fcc','dfcc','hcp'],set_button=0
                       self->my_widget_control,'lattmenu',set_button = (self.flags[2] gt 0),set_value=('Lattice Type'+['',' ('+['BCC','FCC','Diamond FCC','HCP']+')'])[self.flags[2]]
                       recal = 1b
                       end              
         'latt_fcc':   begin
                       self.flags[2] = (self.flags[2] eq 2)?0:2
                       widget_control,event.id,set_button=(self.flags[2] gt 0)
                       self->my_widget_control,'latt_'+['bcc','dfcc','hcp'],set_button=0
                       self->my_widget_control,'lattmenu',set_button = (self.flags[2] gt 0),set_value=('Lattice Type'+['',' ('+['BCC','FCC','Diamond FCC','HCP']+')'])[self.flags[2]]
                       recal = 1b
                       end
         'latt_dfcc':  begin
                       self.flags[2] = (self.flags[2] eq 3)?0:3
                       widget_control,event.id,set_button=(self.flags[2] gt 0)
                       self->my_widget_control,'latt_'+['bcc','fcc','hcp'],set_button=0
                       self->my_widget_control,'lattmenu',set_button = (self.flags[2] gt 0),set_value=('Lattice Type'+['',' ('+['BCC','FCC','Diamond FCC','HCP']+')'])[self.flags[2]]
                       recal = 1b
                       end  
         'latt_hcp':   begin
                       self.flags[2] = (self.flags[2] eq 4)?0:4
                       widget_control,event.id,set_button=(self.flags[2] gt 0)
                       self->my_widget_control,'latt_'+['bcc','fcc','dfcc'],set_button=0
                       self->my_widget_control,'lattmenu',set_button = (self.flags[2] gt 0),set_value=('Lattice Type'+['',' ('+['BCC','FCC','Diamond FCC','HCP']+')'])[self.flags[2]]
                       recal = 1b
                       end
         'ctrlc':      self.plots[self->currenttabnum()]->saveas,'clipboard'                                                          
         'help':       begin
                       info = ['Ghoston reference: H.M. Ronnow et al., Physica B 350 (2004) 11-16.','',$
                               'Enter all fields except for dE, and press Calculate Ghoston button. Scattering to the right refers to clockwise rotation from ki to kf, and '+$
                               'scattering to the left counterclockwise. Clockwise rotation for A3 is positive. When A3 is 0, U is perpendicular to ki.','',$
                               'Check the Options menu for more options.','','Right click over the plot to change its properties; right click over the upper left corner of '+$
                               'the plot to access the plot menu.','','Yiming Qiu (yiming.qiu@nist.gov) 10/2018']
                       ok = dialog_message(info,/center,/info,dialog_parent=event.handler,title='About Ghoston Spurion Viewer')
                       end                                                        
         else:
    endcase
    if keyword_set(recal) and self->check_parm(/nomesg) then begin
       for lefthand=0,1 do self->findhkl,lefthand=lefthand
       for lefthand=0,1 do self->draw,lefthand=lefthand
    endif
end

;cleanup
pro dm_ghoston::Cleanup
    if widget_info(self.tlb,/valid_id) then widget_control,self.tlb,/destroy
    obj_destroy,self.plots
    ptr_free,self.hkl,self.blast,self.listind
    self->save_parm
end

;initialization
function dm_ghoston::Init,lattparm=lattparm,u=u,v=v,workdir=workdir,group_leader=group_leader
    registerName = 'dm_ghoston'
    if xregistered(registerName) then begin ;only allow one copy to be running at one time
       FORWARD_FUNCTION LookupManagedWidget
       id = LookupManagedWidget(registername)
       widget_control,id,/show,/realize,iconify=0
       return,0
    endif
    if self.idl_version ge 5.6 then scr_size = get_screen_size() else device,get_screen_size=scr_size
    if n_elements(group_leader) ne 0 then self.group_leader = group_leader
    self.flags[3:4] = 1b ;default show full circle, show U&V
    self.lattparm[*] = !values.f_nan & self.U[*] = !values.f_nan & self.V[*] = !values.f_nan
    self.Ei = !values.f_nan & self.dEi = !values.f_nan & self.Ef = !values.f_nan & self.dEf = !values.f_nan & self.E = !values.f_nan & self.Q[*] = !values.f_nan
    self.eresl = 0.05
    self.fsize = ([12,13])[scr_size[0] ge 1920]
    self.thick = 1.5
    self.xysize = [47L,5L,750L,500L]
    self.idl_version = dm_to_number(!version.release)
    image_yn = ((self.idl_version ge 8.2) and (!version.os_family eq 'Windows'))
    accel_yn = (self.idl_version ge 6.1)
    
    self->load_parm,/init
    if n_elements(lattparm) eq 6 then self.lattparm = lattparm
    if n_elements(u) eq 3 then self.U = u
    if n_elements(v) eq 3 then self.V = v
    if n_elements(workdir) ne 0 then begin
       if file_test(workdir[0],/directory,/write) then self.workdir = workdir[0]
    endif
    
    bmpfiles = ['mslice_save.bmp']
    defsysv,'!DAVE_AUXILIARY_DIR',exists=exists
    for i=0,n_elements(bmpfiles)-1 do begin
        if exists then bmpfiles[i] = !DAVE_AUXILIARY_DIR+bmpfiles[i] $
        else bmpfiles[i] =  file_which(bmpfiles[i],/include_current_dir)
    endfor
 
    self.tlb = widget_base(title='Ghoston Spurion Viewer',column=1,mbar=mbar,/tlb_size_event,kill_notify='dm_ghoston_Exit',group_leader=group_leader,map=0,uname='tlb',xpad=0)
    filemenu = widget_button(mbar,value='File',/menu)
    optnmenu = widget_button(mbar,value='Options',/menu)
    helpmenu = widget_button(mbar,value='Help',/menu)
    ;file menu
    void     = widget_button(filemenu,value='Save Plot...',uname='save',sensitive=0,_extra=image_yn?{image:dm_resizeicon(bmpfiles[0],/transpose),accelerator:'Ctrl+S'}:(accel_yn?{accelerator:'Ctrl+S'}:{no_copy:1}))
    void     = widget_button(filemenu,value='Print...',uname='print',sensitive=0,_extra=(accel_yn?{accelerator:'Ctrl+P'}:{no_copy:1}))
    void     = widget_button(filemenu,value='Save Parameters...',uname='saveparm',_extra=image_yn?{image:dm_resizeicon(bmpfiles[0],/transpose)}:{no_copy:1},/separator)
    void     = widget_button(filemenu,value='Load Parameters...',uname='loadparm')
    void     = widget_button(filemenu,value='Exit',uname='exit',/separator)
    ;option menu
    lattmenu = dm_widget_button(optnmenu,value='Lattice Type',uname='lattmenu',/menu,set_button=(self.flags[2] gt 0))
    void     = dm_widget_button(lattmenu,value='BCC (H+K+L=even)',uname='latt_bcc',set_button=(self.flags[2] eq 1))
    void     = dm_widget_button(lattmenu,value='FCC (H,K,L all odd or all even)',uname='latt_fcc',set_button=(self.flags[2] eq 2))
    void     = dm_widget_button(lattmenu,value='Diamond FCC (H,K,L all odd, or all even with H+K+L=4n)',uname='latt_dfcc',set_button=(self.flags[2] eq 3))
    void     = dm_widget_button(lattmenu,value='HCP (L even, or H+2K#3n for odd L)',uname='latt_hcp',set_button=(self.flags[2] eq 4))
    
    void     = widget_button(optnmenu,value='Set Default Energy Resolution Factor...',uname='eresl')
    void     = dm_widget_button(optnmenu,value='Isotropic Scaling',uname='isoscale',set_button=self.flags[1],_extra=accel_yn?{accelerator:'Ctrl+I'}:{no_copy:1},/separator)
    void     = dm_widget_button(optnmenu,value='Show Full Ki (Kf) Circle',uname='circle',set_button=self.flags[3])
    void     = dm_widget_button(optnmenu,value='Show U && V',uname='showuv',set_button=self.flags[4])
    void     = widget_button(optnmenu,value='Plot Font Size...',uname='fsize')
    void     = widget_button(optnmenu,value='Plot Line Thickness...',uname='thick')   
    void     = widget_button(optnmenu,value='Copy to Clipboard',uname='ctrlc',_extra=accel_yn?{accelerator:'Ctrl+C'}:{no_copy:1},/separator)
    ;help menu
    void     = widget_button(helpmenu,value='Help',uname='help',_extra=accel_yn?{accelerator:'F1'}:{no_copy:1})
  
    toprow   = widget_base(self.tlb,/row)
    left     = widget_base(toprow,/col,ypad=0,xpad=0,tab_mode=1)
    row      = widget_base(left,/row,ypad=0,xpad=0)
    label    = widget_label(row,value='  Lattice Parameters:')
    geom     = widget_info(label,/geometry)
    symfont  = self->getsymbolfont(2*(fix(geom.scr_ysize)/2))
    
    row      = widget_base(left,/row,ypad=0,xpad=0)
    label    = widget_label(row,value='a ('+string('c5'XB)+')',xsize=self.xysize[0]*1.25,/align_center)
    lattA    = self->my_widget_text(row,value=dm_to_string(self.lattparm[0]),scr_xsize=self.xysize[0],uname='lattA')
    label    = widget_label(row,value='b ('+string('c5'XB)+')',xsize=self.xysize[0],/align_center)
    lattB    = self->my_widget_text(row,value=dm_to_string(self.lattparm[1]),scr_xsize=self.xysize[0],uname='lattB')
    label    = widget_label(row,value='c ('+string('c5'XB)+')',xsize=self.xysize[0],/align_center)
    lattC    = self->my_widget_text(row,value=dm_to_string(self.lattparm[2]),scr_xsize=self.xysize[0],uname='lattC')
    row      = widget_base(left,/row,ypad=0,xpad=0)
    label    = widget_label(row,value='a ('+string('B0'XB)+')',xsize=self.xysize[0]*1.25,/align_center,font=symfont)
    lattAA   = self->my_widget_text(row,value=dm_to_string(self.lattparm[3]),scr_xsize=self.xysize[0],uname='lattAA')
    label    = widget_label(row,value='b ('+string('B0'XB)+')',xsize=self.xysize[0],/align_center,font=symfont)
    lattBB   = self->my_widget_text(row,value=dm_to_string(self.lattparm[4]),scr_xsize=self.xysize[0],uname='lattBB')
    label    = widget_label(row,value='g ('+string('B0'XB)+')',xsize=self.xysize[0],/align_center,font=symfont)
    lattCC   = self->my_widget_text(row,value=dm_to_string(self.lattparm[5]),scr_xsize=self.xysize[0],uname='lattCC')
    row      = widget_base(left,/row,ypad=0,xpad=0)
    label    = widget_label(row,value=' ') 
    row      = widget_base(left,/row,ypad=0,xpad=0)
    fixeief  = dm_widget_droplist(row,xsize=self.xysize[0]*1.25,value='fixed E'+['i','f'],select=self.fixeief,uname='fixEiEf')
    eief     = self->my_widget_text(row,value=dm_to_string(([self.Ei,self.Ef])[self.fixeief]),scr_xsize=self.xysize[0],uname='eief')
    label    = widget_label(row,value=(['dEi','dEf'])[self.fixeief],xsize=self.xysize[0],/align_center,uname='lab_deief')
    deief    = self->my_widget_text(row,value=dm_to_string(([self.dEi,self.dEf])[self.fixeief]),scr_xsize=self.xysize[0],uname='deief')
    label    = widget_label(row,xsize=self.xysize[0],value='E (meV)',/align_center)
    etran    = self->my_widget_text(row,value=dm_to_string(self.E),scr_xsize=self.xysize[0],uname='en')
    row      = widget_base(left,/row,ypad=0,xpad=0)
    label   = widget_label(row,xsize=self.xysize[0]*1.25,value=(['Ef','Ei'])[self.fixeief],/align_center,uname='lab_efei')
    efei    = self->my_widget_text(row,value=dm_to_string(([self.Ef,self.Ei])[self.fixeief]),scr_xsize=self.xysize[0],uname='efei')
    label   = widget_label(row,value=(['dEf','dEi'])[self.fixeief],xsize=self.xysize[0],/align_center,uname='lab_defei')
    defei   = self->my_widget_text(row,value=dm_to_string(([self.dEi,self.dEf])[self.fixeief]),scr_xsize=self.xysize[0],uname='defei')
    right   = widget_base(toprow,/col,ypad=0,xpad=0,tab_mode=1)
    row     = widget_base(right,/row,ypad=0,xpad=0)
    label   = widget_label(row,value='  Scattering Plane:')
    row     = widget_base(right,/row,ypad=0,xpad=0)
    label   = widget_label(row,value='U',xsize=self.xysize[0],/align_center)
    orieUx  = self->my_widget_text(row,value=dm_to_string(self.U[0]),scr_xsize=self.xysize[0],uname='orieUx')
    label   = widget_label(row,value=' ',xsize=self.xysize[1],/align_center)
    orieUy  = self->my_widget_text(row,value=dm_to_string(self.U[1]),scr_xsize=self.xysize[0],uname='orieUy')
    label   = widget_label(row,value=' ',xsize=self.xysize[1],/align_center)
    orieUz  = self->my_widget_text(row,value=dm_to_string(self.U[2]),scr_xsize=self.xysize[0],uname='orieUz')
    row     = widget_base(right,/row,ypad=0,xpad=0)
    label   = widget_label(row,value='V',xsize=self.xysize[0],/align_center)
    orieVx  = self->my_widget_text(row,value=dm_to_string(self.V[0]),scr_xsize=self.xysize[0],uname='orieVx')
    label   = widget_label(row,value=' ',xsize=self.xysize[1],/align_center)
    orieVy  = self->my_widget_text(row,value=dm_to_string(self.V[1]),scr_xsize=self.xysize[0],uname='orieVy')
    label   = widget_label(row,value=' ',xsize=self.xysize[1],/align_center)
    orieVz  = self->my_widget_text(row,value=dm_to_string(self.V[2]),scr_xsize=self.xysize[0],uname='orieVz')
    row     = widget_base(right,/row,ypad=0,xpad=0)
    label   = widget_label(row,value=' ')
    row     = widget_base(right,/row,ypad=0,xpad=0)
    label   = widget_label(row,value='Q',xsize=self.xysize[0],/align_center)
    Qx      = self->my_widget_text(row,value=dm_to_string(self.Q[0]),scr_xsize=self.xysize[0],uname='Qx')
    label   = widget_label(row,value=' ',xsize=self.xysize[1],/align_center)
    orieVy  = self->my_widget_text(row,value=dm_to_string(self.Q[1]),scr_xsize=self.xysize[0],uname='Qy')
    label   = widget_label(row,value=' ',xsize=self.xysize[1],/align_center)
    orieVz  = self->my_widget_text(row,value=dm_to_string(self.Q[2]),scr_xsize=self.xysize[0],uname='Qz')
    row     = widget_base(right,/row,ypad=0,xpad=0)
    label   = widget_label(row,value=' ',scr_xsize=self.xysize[0],/align_center)
    geom    = widget_info(Qx,/geometry)
    calc    = widget_button(row,value='Calculate Ghoston',uname='calc',scr_xsize=self.xysize[0]*3+self.xysize[1]*2+geom.margin*3,frame=1,/align_center)
    if self.idl_version ge 5.6 then widget_control,calc,tooltip='Calculate and plot the ghoston.'
    tabs    = widget_tab(self.tlb,tab_mode=0,uname='tabbase')
    tabr    = widget_base(tabs,title='scattering to right',uname='tab_r',xpad=0,ypad=0,/col)
    tabl    = widget_base(tabs,title='scattering to left',uname='tab_l',xpad=0,ypad=0,/col)
    void    = dm_widget_droplist(tabr,xsize=self.xysize[0]*3.25,value='view all',select=0,uname='viewsel_r',sensitive=0)
    void    = dm_widget_droplist(tabl,xsize=self.xysize[0]*3.25,value='view all',select=0,uname='viewsel_l',sensitive=0)
    geom1   = widget_info(toprow,/geometry)
    xsize   = 800<(fix((scr_size[0]-90)/4*3))>geom1.scr_xsize
    ysize   = fix(2./3*xsize)<((scr_size[1]-100)/2)  
    for i=0,1 do self.plots[i] = obj_new('dm_plot',xsize=xsize,ysize=ysize,widgetbase=([tabr,tabl])[i],background='white',group_leader=self.tlb,/compound,/isolatin1,iso3dtype=1,$
        fsize=self.fsize,ctxtfsize=10,gridontop=0,path=self.workdir)
    dm_center_kid,self.tlb,self.group_leader,/side ;centering the top level base
    widget_control,self.tlb,/realize,/map
    tmp = widget_info(self.tlb,/geometry)
    self.plots[0]->getproperty,xsize=xsize,ysize=ysize
    self.xoffset = tmp.xsize-xsize
    self.yoffset = tmp.ysize-ysize
    tmp = widget_info(toprow,/geometry)
    self.xmin = tmp.xsize
    self.ymin = self.yoffset+fix(self.xmin/3.*2)
    tmp = widget_info(tabs,/geometry)
    self.xmargin = tmp.scr_xsize-xsize
    self.ymargin = tmp.scr_ysize-ysize
    widget_control,self.tlb,set_uvalue=self
    xmanager,registerName,self.tlb,cleanup='dm_ghoston_Exit',/no_block
    return,1
end

;object definition
pro dm_ghoston__define
    void = {dm_ghoston,                 $
            group_leader:  0L,          $   ;group_leader id
            tlb:           0L,          $   ;top level base
            workdir:       '',          $   ;derectory for saving files
            fsize:         12s,         $   ;default fontsize
            thick:         1.5,         $   ;default line thickness
            idl_version:   0e,          $   ;idl version
            txtfocusevent: 0b,          $   ;flag for text focus event
            flags:         bytarr(5),   $   ;misc flags [dE=0.05*E info,isotropic scaling,lattice type,show full circle,show U&V]
            lattparm:      fltarr(6),   $   
            U:             fltarr(3),   $
            V:             fltarr(3),   $
            Q:             fltarr(3),   $
            abc_r:         dblarr(3,3), $
            eresl:         0e,          $   ;default 0.05, dE = eresl*Ei
            fixeief:       0b,          $
            Ei:            0e,          $
            dEi:           0e,          $
            Ef:            0e,          $
            dEf:           0e,          $
            E:             0e,          $
            hkl:           ptrarr(2),   $
            blast:         ptrarr(2),   $
            listind:       ptrarr(2),   $   ;save the index for the viewsel droplist, -1: all, -2: all bragg first  -3:all bragg last 
            kihkl:         fltarr(3,2), $
            kfhkl:         fltarr(3,2), $
            kiran:         fltarr(2),   $
            kfran:         fltarr(2),   $
            a3a4:          fltarr(2,2), $ 
            plots:         objarr(2),   $   ;plot
            xysize:        lonarr(4),   $   ;to record the xsize
            viewsel:       intarr(2),   $   ;
            xoffset:       0L,          $   ;for resize event
            yoffset:       0L,          $   ;for resize event
            xmargin:       0L,          $   ;for resize event
            ymargin:       0L,          $   ;for resize event
            xmin:          0L,          $   ;minimum xsize for resize event
            ymin:          0L           $   ;minimum ysize for resize event
           }
end

pro dm_ghoston,event,lattparm=lattparm,u=u,v=v,workdir=workdir,DAVETool=DAVETool,group_leader=group_leader
    if n_elements(event) ne 0 then begin
       if n_elements(group_leader) eq 0 then group_leader = event.top
       void = obj_new('dm_ghoston',lattparm=lattparm,u=u,v=v,workdir=workdir,group_leader=group_leader) 
    endif else $
       void = obj_new('dm_ghoston',lattparm=lattparm,u=u,v=v,workdir=workdir,group_leader=group_leader)
end



